/* gedcom2ft.c
   Convert a GEDCOM file to 'geneal' format for drawing the tree. 
      Usage: gedc2ft filename synonyms
         will convert the GEDCOM file to 'geneal' format using
         the set of replacement synonyms in the second file,
         and write it to standard output

   Author: Graham Freeman
   Copyright: Graham Freeman, April 1994.
   This program may not be used for commercial purposes without
   the consent of the author.  It may be freely transmitted 
   provided this authorship and copyright notice is retained 
   unmodified.
*/

#include <stdio.h>
#include <string.h>
#define Max(I,J) ((I)>(J)?(I):(J))
#define Min(I,J) ((I)<(J)?(I):(J))

struct wlist {
   char *         text;
   struct wlist * next; };

struct llist {
   void *         item;
   struct llist * next; };

struct indi {
   char *       indiref;
   char *       famcref;
   struct llist famsref;
   char *       name;
   struct wlist comment;
   struct wlist notes;
   struct fam * famc; 
   struct llist fams; };

struct chlist {
   char *          thisref;
   struct indi *   this;
   struct chlist * next; };

struct fam {
   char *        famref;
   char *        husbref;
   char *        wiferef;
   struct indi * husb;
   struct indi * wife;
   struct wlist  comment;
   struct wlist  notes;
   struct chlist child;
   int           level;
   int           multi; };

struct synonym {
   char *word;
   char *syn; };

enum recordtp {Indiv, IndivNote, Family, FamNote, Event, Head};
enum sextp {Male, Female};

char * gettoken(char **ptr, int *len, char **word);
void printancestry(struct fam* this, struct indi* descent);
int settreelevel(struct fam* this);
int printlongestlimb(struct fam* this, struct indi* descent, struct indi* spouse);
void printfamilysummary(struct fam* this, struct indi* descent, struct indi* spouse);
void derefout(char ch[], int nch);
void checktag(char *tag, struct synonym syntbl[], int nsyn, char **word);
int syncompar(const void* first, const void* second);
void *Malloc(long int size);

struct llist     indlist = {NULL,NULL};
struct llist     famlist = {NULL,NULL};
struct llist     evntlist = {NULL,NULL};
struct wlist     headlist = {NULL,NULL};

/**************************************************************/
int main( int argc, char *argv[])
{
   FILE             *src, *syn;
   char *           tagname = "gedcom.tag";
   char             buf[256];
   struct synonym * syntbl;
   int              syntblsiz = 0;
   int              i;
   long int         len, len2;
   int              level, newlevel, prevlevel, skiplevel = -1;
   int              linecount;
   char *           prevkw = NULL;
   struct llist *   curfam = NULL;
   struct llist *   curindi = NULL;
   struct llist *   curevnt = NULL;
   int              nfam = 0, nindi = 0, nevent = 0;
   enum recordtp    currentrec;

   if (argc < 2)
   {
      fprintf(stderr, "Usage: gedc2ft filename tag-file\n"
                      "   will convert the GEDCOM file to 'geneal' format using\n"
                      "   the set of replacement synonyms in the tag-file,\n"
                      "   and write it to standard output\n");
      exit(1);
   }
   if (argc >= 3) tagname = argv[2];
   if ((src = fopen(argv[1],"r")) == NULL)
   {
      fprintf(stderr, "Unable to open the GEDCOM file %s\n", argv[1]);
      exit(1);
   }
   if ((syn = fopen(tagname,"r")) == NULL)
   {
      fprintf(stderr, "Unable to open the synonyms file %s\n", tagname);
      exit(1);
   }

      /* Set up synonym table */
   for (syntblsiz = 0; fgets(buf, sizeof(buf), syn); ++syntblsiz); 
   syntbl = (struct synonym *) Malloc(syntblsiz*sizeof(struct synonym));
   fclose (syn);
   syn = fopen(argv[2],"r");
   for (i=0; fgets(buf, sizeof(buf), syn); ++i )
   {
      buf[strlen(buf)-1] = '\0';
      len = strchr(buf, ' ') - buf;
      if (len <= 0)
      {
         fprintf(stderr,"Badly formed synonym file, line %d: %s\n", i+1, buf);
         exit(1);
      }
      syntbl[i].word = (char *) Malloc(len+1);
      strncpy(syntbl[i].word, buf, len);
      syntbl[i].word[len] = '\0';
      len2 = strspn(&buf[len], " \t");
      syntbl[i].syn = (char *) Malloc(strlen(buf)-len-len2+1);
      strcpy(syntbl[i].syn, &buf[len+len2]);
   }

   level = 0;
   newlevel = 0;
   linecount = 0;
   while (fgets(buf, sizeof(buf), src))
   {
      char *ptr = buf, *word, *tag=NULL, *xref=NULL, *val=NULL;

      ++ linecount;
      buf[strlen(buf)-1] = '\0';
      prevlevel = newlevel;


         /* get the components of a line: level, xref, tag, val */
      if ( !gettoken(&ptr, &len, &word) || sscanf(word,"%d",&newlevel) < 1) 
      {
         fprintf(stderr,"Unable to process line %d of the GEDCOM file: %s\n",
                  linecount,buf);
         exit(1);
      }
      if (newlevel < prevlevel && currentrec == IndivNote) currentrec = Indiv;
      if (newlevel < prevlevel && currentrec == FamNote) currentrec = Family;

      gettoken(&ptr, &len, &word);
      if (word[0] == '@')
      {
         xref = &word[1];
         ptr[-1] = '\0';
         gettoken(&ptr, &len, &word);
      }
      else xref = NULL;
      tag = word;
      if (tag == NULL) continue;
      while (*ptr == ' ' || *ptr == '\t') ++ptr;
      if (*ptr != '\0') val = ptr;
      tag[len] = '\0';


         /* Skip if level is higher than previous one being skipped */
      if (skiplevel >= 0 && newlevel > skiplevel) continue;
      skiplevel = -1;


         /* output previous keyword synonym if necessary */
      if (prevkw != NULL)
      {
         if (newlevel == prevlevel+1)
         {
            if (currentrec == Family)
            {
               struct fam*   this = (struct fam *) curfam->item;
               struct wlist* w = &this->comment;
               if (w->text != NULL)
               {
                  while (w->next != NULL) w = w->next;
                  w->next = (struct wlist *) Malloc(sizeof(struct wlist));
                  w = w->next;
                  w->next = NULL;
               }
               w->text = prevkw;
            }
            else if (currentrec == FamNote)
            {
               struct fam*  this = (struct fam *) curfam->item;
               struct wlist* w = &this->notes;
               if (w->text != NULL)
               {
                  while (w->next != NULL) w = w->next;
                  w->next = (struct wlist *) Malloc(sizeof(struct wlist));
                  w = w->next;
                  w->next = NULL;
               }
               w->text = prevkw;
            }
            else if (currentrec == Indiv)
            {
               struct indi*  this = (struct indi *) curindi->item;
               struct wlist* w = &this->comment;
               if (w->text != NULL)
               {
                  while (w->next != NULL) w = w->next;
                  w->next = (struct wlist *) Malloc(sizeof(struct wlist));
                  w = w->next;
                  w->next = NULL;
               }
               w->text = prevkw;
            }
            else if (currentrec == IndivNote)
            {
               struct indi*  this = (struct indi *) curindi->item;
               struct wlist* w = &this->notes;
               if (w->text != NULL)
               {
                  while (w->next != NULL) w = w->next;
                  w->next = (struct wlist *) Malloc(sizeof(struct wlist));
                  w = w->next;
                  w->next = NULL;
               }
               w->text = prevkw;
            }
            else if (currentrec == Head)
            {
               struct wlist* w = &headlist;
               if (w->text != NULL)
               {
                  while (w->next != NULL) w = w->next;
                  w->next = (struct wlist *) Malloc(sizeof(struct wlist));
                  w = w->next;
                  w->next = NULL;
               }
               w->text = prevkw;
            }
            else if (currentrec == Event)
            {
               free(prevkw);  /* Just discard at present */
            }
         }
         else free(prevkw);
         prevkw = NULL;
      }
      

         /* What is in the synonym table? */
      checktag(tag, syntbl, syntblsiz, &word);
      if (word != NULL && strcmp(word,"-") == 0)
      {
         skiplevel = newlevel;
         continue;
      }


         /* Check against the special tags before just treating as comment */
      if (strcmp(tag,"INDI") == 0)
      {
         struct indi *this = (struct indi *)Malloc(sizeof(struct indi));
         if (indlist.item == NULL) 
         {
            indlist.item = (void *) this;
            curindi = &indlist;
         }
         else
         {
            curindi->next = (struct llist *) Malloc(sizeof(struct llist));
            curindi->next->item = this;
            curindi->next->next = NULL;
            curindi = curindi->next;
         }
         this->indiref = (char *) Malloc(strlen(xref)+1);
         strcpy(this->indiref,xref);
         this->famcref = NULL;
         this->famsref.item = NULL;
         this->famsref.next = NULL;
         this->name = NULL;
         this->comment.text = NULL;
         this->comment.next = NULL;
         this->famc = NULL;
         this->fams.item = NULL;
         this->fams.next = NULL;
         currentrec = Indiv;
         level = newlevel;
         ++ nindi;
      }
      else if (strcmp(tag,"FAM") == 0)
      {
         struct fam *this = (struct fam *)Malloc(sizeof(struct fam));
         if (famlist.item == NULL) 
         {
            famlist.item = (void *) this;
            curfam = &famlist;
         }
         else
         {
            curfam->next = (struct llist *) Malloc(sizeof(struct llist));
            curfam->next->item = this;
            curfam->next->next = NULL;
            curfam = curfam->next;
         }
         this->famref = (char *) Malloc(strlen(xref)+1);
         strcpy(this->famref,xref);
         this->husbref = NULL;
         this->wiferef = NULL;
         this->husb = NULL;
         this->wife = NULL;
         this->comment.text = NULL;
         this->comment.next = NULL;
         this->child.thisref = NULL;
         this->child.this = NULL;
         this->child.next = NULL;
         this->level = 0;
         this->multi = 0;
         currentrec = Family;
         level = newlevel;
         ++ nfam;
      }
      else if (strcmp(tag,"HEAD") == 0)
      {
         currentrec = Head;
         level = newlevel;
      }
      else if (strcmp(tag, "EVEN") == 0)
      {
         fprintf(stderr,
            "Events cannot yet be handled, and will be discarded!\n");
         currentrec = Event;
         level = newlevel;
      }
      else if (strcmp(tag, "NAME") == 0)
      {
         char *  temp;
         int     i;
         if (val == NULL) continue;
         len = strlen(val);
         if (currentrec == Head) len += 3*newlevel;
         temp = (char *) Malloc(len+1);
         temp[0] = '\0';
         if (currentrec == Head) 
            for (i=0; i<newlevel; ++i) strcat(temp,"   ");
         strcat(temp,val);
         if (currentrec == Indiv)
         {
            struct indi * this = (struct indi *) curindi->item;
            if (newlevel == level+1)
            {
               this->name = temp;
            }
            else
            {
               struct wlist* w = &this->comment;
               if (w->text != NULL)
               {
                  while (w->next != NULL) w = w->next;
                  w->next = (struct wlist *) Malloc(sizeof(struct wlist));
                  w = w->next;
                  w->next = NULL;
               }
               w->text = temp;
            }
         }
         else if (currentrec == IndivNote)
         {
            struct indi*  this = (struct indi *) curindi->item;
            struct wlist* w = &this->notes;
            if (w->text != NULL)
            {
               while (w->next != NULL) w = w->next;
               w->next = (struct wlist *) Malloc(sizeof(struct wlist));
               w = w->next;
               w->next = NULL;
            }
            w->text = temp;
         }
         else if (currentrec == Family)
         {
            struct fam*   this = (struct fam *) curfam->item;
            struct wlist* w = &this->comment;
            if (w->text != NULL)
            {
               while (w->next != NULL) w = w->next;
               w->next = (struct wlist *) Malloc(sizeof(struct wlist));
               w = w->next;
               w->next = NULL;
            }
            w->text = temp;
         }
         else if (currentrec == FamNote)
         {
            struct fam*  this = (struct fam *) curfam->item;
            struct wlist* w = &this->notes;
            if (w->text != NULL)
            {
               while (w->next != NULL) w = w->next;
               w->next = (struct wlist *) Malloc(sizeof(struct wlist));
               w = w->next;
               w->next = NULL;
            }
            w->text = temp;
         }
         else if (currentrec == Head)
         {
            struct wlist* w = &headlist;
            if (w->text != NULL)
            {
               while (w->next != NULL) w = w->next;
               w->next = (struct wlist *) Malloc(sizeof(struct wlist));
               w = w->next;
               w->next = NULL;
            }
            w->text = temp;
         }
      }
      else if (strcmp(tag, "FAMS") == 0)
      {
         struct indi * this = (struct indi *) curindi->item;
         struct llist * sref = &this->famsref;
         ++ val;
         val[strlen(val)-1] = '\0';
         if (sref->item != NULL)
         {
            while (sref->next != NULL) sref = sref->next;
            sref->next = (struct llist *) Malloc(sizeof(struct llist));
            sref = sref->next;
            sref->next = NULL;
            sref->item = NULL;
         }
         sref->item = (void *) Malloc(strlen(val)+1);
         strcpy(sref->item,val);
      }
      else if (strcmp(tag, "FAMC") == 0)
      {
         struct indi * this = (struct indi *) curindi->item;
         ++ val;
         val[strlen(val)-1] = '\0';
         this->famcref = (char *) Malloc(strlen(val)+1);
         strcpy(this->famcref,val);
      }
      else if (strcmp(tag, "WIFE") == 0)
      {
         struct fam * this = (struct fam *) curfam->item;
         ++ val;
         val[strlen(val)-1] = '\0';
         this->wiferef = (char *) Malloc(strlen(val)+1);
         strcpy(this->wiferef,val);
      }
      else if (strcmp(tag, "HUSB") == 0)
      {
         struct fam * this = (struct fam *) curfam->item;
         ++ val;
         val[strlen(val)-1] = '\0';
         this->husbref = (char *) Malloc(strlen(val)+1);
         strcpy(this->husbref,val);
      }
      else if (strcmp(tag, "CHIL") == 0)
      {
         struct fam * this = (struct fam *) curfam->item;
         struct chlist *a = &this->child;
         ++ val;
         val[strlen(val)-1] = '\0';
         if (a->thisref != NULL)
         {
            while (a->next != NULL) a = a->next;
            a->next = (struct chlist *) Malloc(sizeof(struct chlist));
            a = a->next;
            a->next = NULL;
            a->this = NULL;
         }
         a->thisref = (char *) Malloc(strlen(val)+1);
         strcpy(a->thisref,val);
      }
      else if (strcmp(tag, "TRLR") == 0)
      {
         break;
      }
      else
         /* process the detail in this line */
      {
         char *  temp;

         if (strcmp(tag,"SUBM") == 0)
         {
            if (newlevel == level) currentrec = Head;
         }
         else if (strcmp(tag,"REPO") == 0)
         {
            if (newlevel == level) currentrec = Head;
         }
         else if (strcmp(tag,"SOUR") == 0)
         {
            if (newlevel == level) currentrec = Head;
         }
         else if (strcmp(tag, "NOTE") == 0 && currentrec == Indiv)
         {
            currentrec = IndivNote;
         }
         else if (strcmp(tag, "NOTE") == 0 && currentrec == Family)
         {
            currentrec = FamNote;
         }

         if (word == NULL || strlen(word) == 0)
         {
            int i;
            if (val == NULL) continue;
            len = strlen(val)+2;
            if (xref != NULL) len += strlen(xref) + 3;
            if (currentrec == Head) len += 3*newlevel;
            temp = (char *) Malloc(len+1);
            temp[0] = '\0';
            if (currentrec == Head) 
               for (i=0; i<newlevel; ++i) strcat(temp,"   ");
            strcat(temp,val);
            if (xref != NULL) 
            {
               strcat(temp," @");
               strcat(temp,xref);
               strcat(temp,"@");
            }
            strcat(temp,"  ");
         }
         else if (strcmp(word,"-") == 0) continue;
         else 
         {
            int i;
            len = 0;
            if (val != NULL) len = strlen(val)+2;
            if (xref != NULL) len += strlen(xref) + 3;
            if (currentrec == Head) len += 3*newlevel;
            len2 = strlen(word);
            temp = (char *) Malloc(len+len2+2);
            temp[0] = '\0';
            if (currentrec == Head) 
               for (i=0; i<newlevel; ++i) strcat(temp,"   ");
            strcat(temp,word);
            if (xref != NULL) 
            {
               strcat(temp," @");
               strcat(temp,xref);
               strcat(temp,"@");
            }
            strcat(temp," ");
            if (val == NULL)
            {
               prevkw = temp;
               continue;
            }
            strcat(temp,val);
            strcat(temp,"  ");
         }
         if (currentrec == Family)
         {
            struct fam*   this = (struct fam *) curfam->item;
            struct wlist* w = &this->comment;
            if (w->text != NULL)
            {
               while (w->next != NULL) w = w->next;
               w->next = (struct wlist *) Malloc(sizeof(struct wlist));
               w = w->next;
               w->next = NULL;
            }
            w->text = temp;
         }
         else if (currentrec == Indiv)
         {
            struct indi*  this = (struct indi *) curindi->item;
            struct wlist* w = &this->comment;
            if (w->text != NULL)
            {
               while (w->next != NULL) w = w->next;
               w->next = (struct wlist *) Malloc(sizeof(struct wlist));
               w = w->next;
               w->next = NULL;
            }
            w->text = temp;
         }
         else if (currentrec == IndivNote)
         {
            struct indi*  this = (struct indi *) curindi->item;
            struct wlist* w = &this->notes;
            if (w->text != NULL)
            {
               while (w->next != NULL) w = w->next;
               if (strcmp(tag, "NOTE") == 0 && w != &this->notes)
               {
                  char *b;
                  w->next = (struct wlist *) Malloc(sizeof(struct wlist));
                  w = w->next;
                  w->next = NULL;
                  b = (char *) Malloc(2);
                  b[0] = '\0';
                  b[1] = '\0';
                  w->text = b;
               }
               w->next = (struct wlist *) Malloc(sizeof(struct wlist));
               w = w->next;
               w->next = NULL;
            }
            w->text = temp;
         }
         else if (currentrec == FamNote)
         {
            struct fam*  this = (struct fam *) curfam->item;
            struct wlist* w = &this->notes;
            if (w->text != NULL)
            {
               while (w->next != NULL) w = w->next;
               if (strcmp(tag, "NOTE") == 0 && w != &this->notes)
               {
                  char *b;
                  w->next = (struct wlist *) Malloc(sizeof(struct wlist));
                  w = w->next;
                  w->next = NULL;
                  b = (char *) Malloc(2);
                  b[0] = '\0';
                  b[1] = '\0';
                  w->text = b;
               }
               w->next = (struct wlist *) Malloc(sizeof(struct wlist));
               w = w->next;
               w->next = NULL;
            }
            w->text = temp;
         }
         else if (currentrec == Head)
         {
            struct wlist* w = &headlist;
            if (w->text != NULL)
            {
               while (w->next != NULL) w = w->next;
               w->next = (struct wlist *) Malloc(sizeof(struct wlist));
               w = w->next;
               w->next = NULL;
            }
            w->text = temp;
         }
         else if (currentrec == Event)
         {
            free(temp);  /* Just discard at present */
         }
      }
   }


      /* process the references to form linkages to data items */
      /*   first: references to individuals in families */
   for (curfam = &famlist; curfam != NULL; curfam = curfam->next)
   {
      struct fam*   this = (struct fam *) curfam->item;
      if (this == NULL)
      {
         fprintf(stderr,"No families found.  Tree cannot be built\n");
         break;
      }
      if (this->husbref != NULL)
      {
         struct llist* il = &indlist;
         while (il != NULL)
         {
            struct indi*  p = (struct indi *) il->item;
            if (strcmp(p->indiref,this->husbref) == 0)
            {
               this->husb = p;
               break;
            }
            il = il->next;
         }
         if (il == NULL) 
            fprintf(stderr,"Reference to %s in family %s not found!\n",
                  this->husbref, this->famref);
      }
      if (this->wiferef != NULL)
      {
         struct llist* il = &indlist;
         for (;il != NULL; il = il->next)
         {
            struct indi*  p = (struct indi *) il->item;
            if (strcmp(p->indiref,this->wiferef) == 0)
            {
               this->wife = p;
               break;
            }
         }
         if (il == NULL)
            fprintf(stderr,"Reference to %s in family %s not found!\n",
                  this->wiferef, this->famref);
      }
      if (this->child.thisref != NULL)
      {
         struct chlist *c = &this->child;
         for (; c != NULL; c = c->next)
         {
            struct llist* il = &indlist;
            for (;il != NULL; il = il->next)
            {
               struct indi*  p = (struct indi *) il->item;
               if (strcmp(p->indiref,c->thisref) == 0)
               {
                  c->this = p;
                  break;
               }
            }
            if (il == NULL)
               fprintf(stderr,"Reference to %s in family %s not found!\n",
                     c->thisref, this->famref);
         }
      }
   }
      /* second: family references for all individuals */
   for (curindi = &indlist; curindi != NULL; curindi = curindi->next)
   {
      struct indi*   this = (struct indi *) curindi->item;
      if (this->famcref != NULL)
      {
         struct llist* fl = &famlist;
         for (;fl != NULL; fl = fl->next)
         {
            struct fam*  f = (struct fam *) fl->item;
            if (strcmp(f->famref,this->famcref) == 0)
            {
               this->famc = f;
               break;
            }
         }
         if (fl == NULL)
            fprintf(stderr,"Reference to family %s in %s not found!\n",
                  this->famcref, this->indiref);
      }
      if (this->famsref.item != NULL)
      {
         struct llist* sl = &this->famsref;
         for (; sl != NULL; sl = sl->next)
         {
            struct llist* fl = &famlist;
            for (;fl != NULL; fl = fl->next)
            {
               struct fam*  f = (struct fam *) fl->item;
               if (strcmp(f->famref,sl->item) == 0)
               {
                  struct llist * s = &this->fams;
                  if (s->item != NULL)
                  {
                     while (s->next != NULL) s = s->next;
                     s->next = (struct llist *) Malloc(sizeof(struct llist));
                     s = s->next;
                     s->next = NULL;
                     s->item = NULL;
                  }
                  s->item = (void *) f;
                  break;
               }
            }
            if (fl == NULL)
               fprintf(stderr,"Reference to family %s in %s not found!\n",
                     sl->item, this->indiref);
         }
      }
   }
      /* third: any references in events? */
   for (curevnt = &evntlist; curevnt != NULL; curevnt = curevnt->next)
   {
   }

      /* output any HEAD information */
   {
      struct wlist* w = &headlist;
      if (w->text != NULL)
      {
         printf("%%NF%%");
         for (; w != NULL; w = w->next)
         {
            derefout(w->text, strlen(w->text));
            putchar('\n');
         }
         printf("%%FI%%%%NP%%\n");
      }
   }
   
      /* output each family */
      /* find oldest source families (ones without ancestors) */
   for (curfam = &famlist; curfam != NULL; curfam = curfam->next)
   {
      struct fam*   this = (struct fam *) curfam->item;
      int           husblevel=0, wifelevel=0;
      if (this == NULL) break;
      if (this->husb != NULL && this->husb->famc != NULL) 
         husblevel = settreelevel(this->husb->famc);
      if (this->wife != NULL && this->wife->famc != NULL) 
         wifelevel = settreelevel(this->wife->famc);
      if (husblevel >= wifelevel)
      {
         if (this->husb != NULL && this->husb->famc != NULL) 
            printancestry(this->husb->famc,this->husb);
         if (this->wife != NULL && this->wife->famc != NULL) 
            printancestry(this->wife->famc,this->wife);
      }
      else
      {
         if (this->wife != NULL && this->wife->famc != NULL) 
            printancestry(this->wife->famc,this->wife);
         if (this->husb != NULL && this->husb->famc != NULL) 
            printancestry(this->husb->famc,this->husb);
      }
      printf("<generation>\n");
      if (this->husb != NULL)
      {
         printf("%s",this->husb->name);
         if (this->husb->comment.text != NULL)
         {
            struct wlist* w = &this->husb->comment;
            printf("[");
            while (w != NULL)
            {
               int n = strlen(w->text);
               while (w->text[n-1] == ' ' && n > 0) --n;
               derefout(w->text, n);
               if (w->next != NULL) printf("//");
               w = w->next;
            }
            printf("]");
         }
      }
      if (this->wife != NULL)
      {
         if (this->husb != NULL) 
         {
            printf("=");
            if (this->comment.text != NULL)
            {
               struct wlist* w = &this->comment;
               putchar('[');
               while (w != NULL)
               {
                  derefout(w->text,strlen(w->text));
                  printf("//");
                  w = w->next;
               }
               putchar(']');
            }
         }
         printf("%s",this->wife->name);
         if (this->wife->comment.text != NULL)
         {
            struct wlist* w = &this->wife->comment;
            printf("[");
            while (w != NULL)
            {
               int n = strlen(w->text);
               while (w->text[n-1] == ' ' && n > 0) --n;
               derefout(w->text, n);
               if (w->next != NULL) printf("//");
               w = w->next;
            }
            printf("]");
         }
      }
      printf("\n");
      if (this->child.this != NULL)
      {
         struct chlist* p = &this->child;
         printf("<children>\n");
         while (p != NULL)
         {
            printf("<>%s\n",p->this->name);
            if (p->this->comment.text != NULL)
            {
               struct wlist* w = &p->this->comment;
               while (w != NULL)
               {
                  derefout(w->text,strlen(w->text));
                  w = w->next;
               }
               printf("\n");
            }
            p = p->next;
         }
         printf("<nomorechildren>\n");
      }
      printf("<comment>\n");
      if (this->notes.text != NULL)
      {
         struct wlist* w = &this->comment;
         while (w != NULL)
         {
            derefout(w->text,strlen(w->text));
            putchar('\n');
            w = w->next;
         }
      }
      if (this->husb != NULL && this->husb->notes.text != NULL)
      {
         struct wlist* w = &this->husb->notes;
         if (this->comment.text != NULL) printf("\n");
         printf("%s%%L%%\n",this->husb->name);
         while (w != NULL)
         {
            derefout(w->text,strlen(w->text));
            printf("\n");
            w = w->next;
         }
      }
      if (this->wife != NULL && this->wife->comment.text != NULL)
      {
         struct wlist* w = &this->wife->notes;
         if (this->husb != NULL && this->husb->notes.text != NULL)
            printf("\n");
         printf("%s%%L%%\n",this->wife->name);
         while (w != NULL)
         {
            derefout(w->text,strlen(w->text));
            printf("\n");
            w = w->next;
         }
      }
      if (this->child.this != NULL)
      {
         struct chlist* p = &this->child;
         while (p != NULL)
         {
            if (p->this->fams.item == NULL && p->this->notes.text != NULL)
            {
               struct wlist* w = &p->this->notes;
               printf("\n%s%%L%%\n",p->this->name);
               while (w != NULL)
               {
                  derefout(w->text,strlen(w->text));
                  printf("\n");
                  w = w->next;
               }
            }
            p = p->next;
         }
      }
   }
}

/**************************************************************/
char * gettoken(char **ptr, int *len, char **word)
   /* Given the address of a pointer to the text string, find the
      address of the start of the next word and its length.
      Replace the starting address with the end address so that
      subsequent calls will not give the same string. */
{
   while (**ptr == ' ' || **ptr == '\t') ++(*ptr);
   *len = 0;
   if (**ptr == '\0') 
   {
      *word = NULL;
   }
   else 
   {
      *word = *ptr;
      while (**ptr != ' ' && **ptr != '\t' && **ptr != '\0') 
      {
         ++(*ptr);
         ++(*len);
      }
   }
   return(*word);
}

/**************************************************************/
void printancestry(struct fam* this, struct indi* descent)
   /* Use recursive routines to go back through the ancestry of this 
      family to its sources and generate <ancestry> sections in the 
      output, in order of length of descent line, longest first. 
      The person 'descent' is the one we are interested in, so this
      name should be highlighted. */ 
{
   while (this->level != 0) 
   {
      printlongestlimb(this,descent,NULL);
   }
}

/**************************************************************/
void cleartreelevel(struct fam* this)
   /* Recursively proceed through the tree, clearing the 'level' of each
      family and the 'multi' count that handles linked ancestry. */
{
   int motherlevel, fatherlevel;
   this->level = 0;
   this->multi = 0;
   if (this->husb != NULL && this->husb->famc != NULL) 
      cleartreelevel(this->husb->famc);
   if (this->wife != NULL && this->wife->famc != NULL) 
      cleartreelevel(this->wife->famc);
}

/**************************************************************/
int settreelevel(struct fam* this)
   /* Recursively proceed through the tree, setting the 'level' of each
      family to the distance that node in the tree is from the oldest
      ancestor along that branch.  The oldest will have level 1. */
{
   int motherlevel, fatherlevel;
   if (this->husb != NULL && this->husb->famc != NULL) 
      fatherlevel = settreelevel(this->husb->famc);
   else fatherlevel = 0;
   if (this->wife != NULL && this->wife->famc != NULL) 
      motherlevel = settreelevel(this->wife->famc);
   else motherlevel = 0;
   this->level = 1 + ((fatherlevel > motherlevel) ? fatherlevel : motherlevel);
   ++ this->multi;
   return(this->level);
}

/**************************************************************/
int printlongestlimb(
                  struct fam*  this, 
                  struct indi* descent, 
                  struct indi* spouse)
   /* Using level numbers set in the family structures, recursively
      process the longest limb of the tree, printing in order from
      oldest to youngest.  Prune the processed parts by setting
      level numbers to indicate which parts remain to be processed,
      the depth of those parts, and the route, by setting level
      to zero if no parts remain in that branch, or negative depth
      where a node has been processed but a branch has not. 
      The individual 'descent' is to be highlighted in the tree, along
      with others in the descent line to this person.  This person's
      spouse is to be placed in the ancestry section beside the name. 

      The algorithm used here will not work in all situations, in
      particular when there is more than one line of descent from
      an individual.  The 'multi' item has been added to handle 
      the most common multiple path from an ancestor, but some
      obscure cases will have links missing, eg
                1
                |
                2     3
               / \   /
              4    5
             /      \
            6        7
             \     /
              8 = 9
      will print 1-2-4-6-8 on the first call, and 3-5-7-9 on the second,
      with nothing further printed, and the link 2-5 implied by the names 
      without being explicitly mentioned.  However, the tree
                1
                |
                2 
               / \
              4    5
             /      \
            6        7
             \     /
              8 = 9
      will correctly print 1-2-4-6-8 on the first call, and 2-5-7-9
      on the second.
   */
{
   int motherlevel, fatherlevel;

   if (this->level != 0)
   {
      if (this->husb != NULL && this->husb->famc != NULL) 
         fatherlevel = this->husb->famc->level;
      else fatherlevel = 0;
      if (this->wife != NULL && this->wife->famc != NULL) 
         motherlevel = this->wife->famc->level;
      else motherlevel = 0;

      if (fatherlevel !=0 && abs(fatherlevel) >= abs(motherlevel))
         printlongestlimb(this->husb->famc,this->husb,this->wife);
      else if (motherlevel != 0) 
         printlongestlimb(this->wife->famc,this->wife,this->husb);
      else if (this->level > 1)
      {
         if (this->husb != NULL && this->husb->famc != NULL && 
             this->husb->famc->multi > 0) 
            printlongestlimb(this->husb->famc,this->husb,this->wife);
         else if (this->wife != NULL && this->wife->famc != NULL && 
             this->wife->famc->multi > 0) 
            printlongestlimb(this->wife->famc,this->wife,this->husb);
      }
   }

   if (this->level >= 0)
      printfamilysummary(this, descent, spouse);

   if (this->husb != NULL && this->husb->famc != NULL) 
      fatherlevel = this->husb->famc->level;
   else 
      fatherlevel = 0;
   if (this->wife != NULL && this->wife->famc != NULL) 
      motherlevel = this->wife->famc->level;
   else 
      motherlevel = 0;
   if (motherlevel == 0 && fatherlevel == 0) 
       this->level = 0;
   else
      this->level = -1-Max(abs(motherlevel),abs(fatherlevel));
   --this->multi;
}

/**************************************************************/
void printfamilysummary(
                  struct fam*  this, 
                  struct indi* descent, 
                  struct indi* spouse)
   /* Print <ancestry>, the parents, and the children, with one
      marked as special, the line of descent being pursued. */
{
   if (this->level == 1 || this->level == 0)
   {
      printf("<ancestry>\n");
      if (this->husb != NULL) printf("%s",this->husb->name);
      if (this->wife != NULL)
      {
         if (this->husb != NULL) printf("=");
         printf("%s",this->wife->name);
      }
      printf("\n");
   }
   printf("<children>\n");
   {
      struct chlist* p = &this->child;
      while (p != NULL)
      {
         if (p->this == descent) 
         {
            printf("<*>%s",p->this->name);
            if (spouse != NULL) printf(" = %s",spouse->name);
            printf("\n");
         }
         else printf("<>%s\n",p->this->name);
         p = p->next;
      }
   }
}

/**************************************************************/
void derefout(char ch[], int nch)
   /* Output the character array, checking first for references which
      should be checked against the list of individuals, with their name
      output instead. */
{
   char ref[50];
   int i;
   for (i=0; i<nch; ++i)
   {
      if (ch[i] == '@')
      {
         int j=i+1;
         for (j=i+1; ch[j] != '\0' && ch[j] != '@'; ++j) 
            ref[j-i-1] = ch[j];
         if (ch[j] == '@')
         {
            if (j == i+1)
            {
               putchar('@');
               i = j;
            }
            else
            {
               struct llist* il = &indlist;
               ref[j-i-1] = '\0';
               while (il != NULL)
               {
                  struct indi*  p = (struct indi *) il->item;
                  if (strcmp(p->indiref,ref) == 0)
                  {
                     printf("%s", p->name);
                     break;
                  }
                  il = il->next;
               }
               if (il == NULL) 
                  printf("@%s@",ref);
               i = j;
            }
         }
         else
         {
            j = j-i-1;
            for (i=0; i<j; ++i)
               putchar(ref[i]);
            return;
         }
      }
      else putchar(ch[i]);
   }
}

/**************************************************************/
void checktag(char *tag, struct synonym syntbl[], int nsyn, char **word)
   /* Search through the synonym table for one matching the given tag.
      Return in 'word' the address of the replacement text, or NULL. */
{
   struct synonym sought;
   struct synonym *result;
   sought.word = tag;
   result = (struct synonym *)
         bsearch(&sought,syntbl,nsyn,sizeof(struct synonym),syncompar);
   if (result == NULL) *word = NULL;
   else                *word = result->syn;
}

/**************************************************************/
int syncompar(const void* first, const void* second)
   /* comparison routine used by bsearch in comparing synonyms */
{
   struct synonym *a = (struct synonym*)first, *b = (struct synonym*)second;
   return(strcmp(a->word, b->word));
}

/**************************************************************/
void *Malloc(long int size)
{
   void *a = (void *) malloc(size);
   if (a == NULL) 
   {
      fprintf(stderr,"Memory allocation failure, trying to get %ld bytes\n",
            size);
      size = size / (int) a; /* cause dump */
      exit(1);
   }
   return(a);
}
